from multiprocessing import Value
from pyathena import connect
from pyathena.util import as_pandas

from red_panda.typing import AthenaQueryResult
from red_panda.aws import AWSUtils

import logging


LOGGER = logging.getLogger(__name__)


class AthenaUtils(AWSUtils):
    """AWS Athena operations.

    Args:
        aws_config: AWS configuration.
        s3_staging_dir: Full S3 folder uri, i.e. s3://athena-query/results.
        region_name: AWS region name.

    Attributes:
        aws_config: AWS configuration.
        s3_staging_dir: Full S3 folder uri, i.e. s3://athena-query/results.
        region_name: AWS region name.

    TODO:
        * Complete Support for other cursor types.
        * Full parameters on `connect`.
    """

    def __init__(
        self,
        aws_config: dict,
        s3_staging_dir: dict,
        work_group: str = None,
        region_name: str = None,
    ):
        super().__init__(aws_config=aws_config)
        self.work_group = work_group
        self.cursor = connect(
            aws_access_key_id=self.aws_config.get("aws_access_key_id"),
            aws_secret_access_key=self.aws_config.get("aws_secret_access_key"),
            s3_staging_dir=s3_staging_dir,
            work_group=work_group,
            region_name=region_name or aws_config.get("region_name"),
        ).cursor()

    def run_query(
        self, sql: str, as_df: bool = False, use_cache: bool = False
    ) -> AthenaQueryResult:
        """Run query on Athena.

        Args:
            sql: SQL query.
            as_df (optional): Whether to return the result as DataFrame.

        Returns:
            Query result as DataFrame or List[Tuple].
        """
        cache_size = 0
        if use_cache:
            if self.work_group is None:
                raise ValueError("Workgroup must be specified to use cache.")
            cache_size = 50
        self.cursor.execute(sql, cache_size=cache_size)

        if as_df:
            return as_pandas(self.cursor)

        res = []
        desc = self.cursor.description
        for row in self.cursor:
            r = {}
            for i, c in enumerate(desc):
                r[c[0]] = row[i]
            res.append(r)
        return res



import warnings
from collections import OrderedDict
from textwrap import dedent
from typing import Union, Optional, List
import logging

import pandas as pd

from red_panda.pandas import PANDAS_TOCSV_KWARGS
from red_panda.aws import (
    REDSHIFT_RESERVED_WORDS,
    REDSHIFT_COPY_KWARGS,
)
from red_panda.utils import filter_kwargs, make_valid_uri
from red_panda.aws.s3 import S3Utils
from red_panda.aws.redshift import RedshiftUtils


LOGGER = logging.getLogger(__name__)


def map_types(columns_types: dict) -> dict:
    """Convert Pandas dtypes to Redshift data types.

    Args:
        cols_types: The return value of `dict(df.dtypes)`, where `df` is a Pandas Dataframe.

    Returns:
        dict: A `dict` of `{original column name: mapped redshift data type}`
    """
    PANDAS_REDSHIFT_TYPES_MAP = {
        "O": "varchar(256)",
        "object": "varchar(256)",
        "int64": "bigint",
        "float64": "double precision",
        "boolean": "boolean",
    }
    return {
        c: {"data_type": PANDAS_REDSHIFT_TYPES_MAP.get(t.name)}
        if PANDAS_REDSHIFT_TYPES_MAP.get(t.name) is not None
        else {"data_type": "varchar(256)"}
        for c, t in columns_types.items()
    }


def check_invalid_columns(columns: list):
    """Check column names against Redshift reserved words.

    Args:
        columns: A list of column names to check.

    Raises:
        ValueError: If the column name is invalid.
    """
    invalid_df_col_names = [c for c in columns if c in REDSHIFT_RESERVED_WORDS]
    if len(invalid_df_col_names) > 0:
        raise ValueError(f"Redshift reserved words: {invalid_df_col_names}")


class RedPanda(RedshiftUtils, S3Utils):
    """Class for operations between Pandas and Redshift/S3.

    Args:
        redshift_config: Redshift configuration.
        aws_config (optional): AWS configuration.
        default_bucket (optional): Default bucket to store files.
        dryrun (optional): If True, queries will be printed instead of executed.
    
    Attributes:
        redshift_config (dict): Redshift configuration.
        aws_config (dict): AWS configuration.
        default_bucket (str): Default bucket to store files.
    """

    def __init__(
        self,
        redshift_config: dict,
        aws_config: dict,
        default_bucket: str = None,
        dryrun: bool = False,
    ):
        RedshiftUtils.__init__(self, redshift_config, dryrun)
        S3Utils.__init__(self, aws_config)
        self.default_bucket = default_bucket

    def s3_to_redshift(
        self,
        bucket: str,
        key: str,
        table_name: str,
        column_definition: dict = None,
        append: bool = False,
        delimiter: str = ",",
        ignoreheader: int = 1,
        quote_character: str = '"',
        dateformat: str = "auto",
        timeformat: str = "auto",
        acceptinvchars: str = "?",
        acceptanydate: bool = False,
        blanksasnull: bool = False,
        emptyasnull: bool = False,
        escape: bool = False,
        null: str = None,
        encoding: str = None,
        explicit_ids: bool = False,
        fillrecord: bool = False,
        ignoreblanklines: bool = False,
        removequotes: bool = False,
        roundec: bool = False,
        trimblanks: bool = False,
        truncatecolumns: bool = False,
        region: str = None,
        iam_role: str = None,
        column_list: list = None,
    ):
        """Load S3 file into Redshift.

        Args:
            bucket: S3 bucket name.
            key: S3 key.
            table_name: Redshift table name (optional include schema name).
            column_definition (optional): Specify the column definition if for CREATE TABLE.
            append (optional): Ff True, df will be appended to Redshift table, otherwise table will 
                be dropped and recreated.
            delimiter (optional): Delimiter of file. Default is ",".
            ignoreheader (optional): number of header lines to skip when COPY.
            quote_character (optional): QUOTE_CHARACTER for COPY. Only used when delimiter is ",". 
                Default to '"'.
            dateformat (optional): TIMEFORMAT argument for COPY. Default is "auto". 
            timeformat (optional): TIMEFORMAT argument for COPY. Default is "auto".
            acceptinvchars (optional): Whether to include the ACCEPTINVCHAR argument in COPY.
            acceptanydate (optional): Allows any date format, including invalid formats.
            blanksasnull (optional): Loads blank fields, which consist of only white space 
                characters, as NULL.
            emptyasnull (optional): Indicates that Amazon Redshift should load empty CHAR and 
                VARCHAR fields as NULL.
            escape (optional): Whether to include the ESCAPE argument in COPY.
            null (optional): Specify the NULL AS string.
            encoding (optional): Specifies the encoding type of the load data.
            explicit_ids (optional): Use EXPLICIT_IDS with tables that have IDENTITY column.
            fillrecord (optional): Allows data files to be loaded when contiguous columns are 
                missing at the end of some of the records.
            ignoreblanklines (optional): Ignores blank lines that only contain a line feed in a data
                file and does not try to load them.
            removequotes (optional): Removes surrounding quotation marks from strings in the 
                incoming data.
            roundec (optional): Rounds up numeric values when the scale of the input value is 
                greater than the scale of the column. 
            trimblanks (optional): Removes the trailing white space characters from a VARCHAR 
                string.
            truncatecolumns (optional): Truncates data in columns to the appropriate number of 
                characters so that it fits the column specification.
            region (optional): S3 region.
            iam_role (optional): Use IAM Role for access control.
            column_list (optional): List of columns to COPY.
        
        TODO:
            * Handle S3 client side encryption.
            * Handle COPY using manifest file.
        """
        if not append:
            if column_definition is None:
                raise ValueError("column_definition cannot be None if append is False")
            else:
                drop_first = False if append else True
                self.create_table(table_name, column_definition, drop_first=drop_first)

        s3_source = f"s3://{bucket}/{key}"
        quote_option = (
            f"csv quote as '{quote_character}'"
            if delimiter == "," and not escape
            else ""
        )
        region_option = f"region '{region}'" if region is not None else ""
        escape_option = "escape" if escape else ""
        acceptinvchars_option = (
            f"acceptinvchars as '{acceptinvchars}'"
            if acceptinvchars is not None
            else ""
        )
        acceptanydate_option = "acceptanydate" if acceptanydate else ""
        blanksasnull_option = "blanksasnull" if blanksasnull else ""
        emptyasnull_option = "emptyasnull" if emptyasnull else ""
        explicit_ids_option = "explicit_ids" if explicit_ids else ""
        fillrecord_option = "fillrecord" if fillrecord else ""
        ignoreblanklines_option = "ignoreblanklines" if ignoreblanklines else ""
        removequotes_option = "removequotes" if removequotes else ""
        roundec_option = "roundec" if roundec else ""
        trimblanks_option = "trimblanks" if trimblanks else ""
        truncatecolumns_option = "truncatecolumns" if truncatecolumns else ""
        encoding_option = f"encoding as {encoding}" if encoding is not None else ""
        null_option = f"null as '{null}'" if null is not None else ""
        aws_access_key_id = self.aws_config.get("aws_access_key_id")
        aws_secret_access_key = self.aws_config.get("aws_secret_access_key")
        if (
            aws_access_key_id is None
            and aws_secret_access_key is None
            and iam_role is None
        ):
            raise ValueError(
                "Must provide at least one of [iam_role, aws_access_key_id/aws_secret_access_key]"
            )
        aws_token = self.aws_config.get("aws_session_token")
        aws_token_option = (
            f"session_token '{aws_token}'" if aws_token is not None else ""
        )
        if iam_role is not None:
            iam_role_option = f"iam_role '{iam_role}'"
            access_key_id_option = ""
            secret_access_key_option = ""
        else:
            iam_role_option = ""
            access_key_id_option = f"access_key_id '{aws_access_key_id}'"
            secret_access_key_option = f"secret_access_key '{aws_secret_access_key}'"
        column_list_option = ""
        if column_list is not None:
            column_list_option = f"({','.join(column_list)})"
        copy_template = f"""\
        copy {table_name} {column_list_option}
        from '{s3_source}' 
        delimiter '{delimiter}'
        {quote_option}
        {escape_option}
        {acceptinvchars_option}
        {acceptanydate_option}
        {blanksasnull_option}
        {emptyasnull_option}
        {null_option}
        {encoding_option}
        {explicit_ids_option}
        {fillrecord_option}
        {removequotes_option}
        {roundec_option}
        {trimblanks_option}
        {truncatecolumns_option}
        {ignoreblanklines_option}
        ignoreheader {ignoreheader}
        dateformat '{dateformat}'
        timeformat '{timeformat}'
        {access_key_id_option}
        {secret_access_key_option}
        {aws_token_option}
        {iam_role_option}
        {region_option}
        """
        self.run_query(copy_template)

    def df_to_redshift(
        self,
        df: pd.DataFrame,
        table_name: str,
        bucket: str = None,
        column_definition: dict = None,
        append: bool = False,
        path: str = None,
        file_name: str = None,
        cleanup: bool = True,
        **kwargs,
    ):
        """Pandas DataFrame to Redshift table.

        Args:
            df: Source dataframe.
            table_name: Redshift table name (optionally include schema name).
            bucket (optional): S3 bucket name, fallback to `default_bucket` if not present.
            column_definition (optional): Specify the column definition for CREATE TABLE. If 
                not given and append is False, data type will be inferred.
            append (optional): If true, df will be appended to Redshift table, otherwise table 
                will be dropped and recreated.
            path (optional): S3 key excluding file name.
            file_name (optional): If None, file_name will be randomly generated.
            cleanup: (optional): Default True, S3 file will be deleted after COPY.
            **kwargs: keyword arguments to pass to Pandas `to_csv` and Redshift COPY.
        """
        bridge_bucket = bucket or self.default_bucket
        if not bridge_bucket:
            raise ValueError("Either bucket or default_bucket must be provided.")

        to_csv_kwargs = filter_kwargs(kwargs, PANDAS_TOCSV_KWARGS)
        copy_kwargs = filter_kwargs(kwargs, REDSHIFT_COPY_KWARGS)

        if column_definition is None:
            column_definition = map_types(OrderedDict(df.dtypes))

        # default pandas behavior is true when index is not specified
        if to_csv_kwargs.get("index") is None or to_csv_kwargs.get("index"):
            if df.index.name:
                full_column_definition = OrderedDict({df.index.name: df.index.dtype})
            else:
                full_column_definition = OrderedDict({"index": df.index.dtype})
            full_column_definition = map_types(full_column_definition)
            full_column_definition.update(column_definition)
            column_definition = full_column_definition

        check_invalid_columns(list(column_definition))

        if file_name is None:
            import uuid

            file_name = f"redpanda-{uuid.uuid4()}"

        s3_key = make_valid_uri(path if path is not None else "", file_name)
        self.df_to_s3(df, bucket=bridge_bucket, key=s3_key, **to_csv_kwargs)
        try:
            self.s3_to_redshift(
                bridge_bucket,
                s3_key,
                table_name,
                column_definition=column_definition,
                append=append,
                **copy_kwargs,
            )
        finally:
            if cleanup:
                self.delete_from_s3(bridge_bucket, s3_key)

    def redshift_to_s3(
        self,
        sql: str,
        bucket: str = None,
        path: str = None,
        prefix: str = None,
        iam_role: str = None,
        file_format: str = None,
        partition_by: List[str] = None,
        include_partition_column: bool = False,
        manifest: bool = False,
        header: bool = False,
        delimiter: str = None,
        fixedwidth: Union[str, int] = None,
        encrypted: bool = False,
        bzip2: bool = False,
        gzip: bool = False,
        zstd: bool = False,
        addquotes: bool = False,
        null: str = None,
        escape: bool = False,
        allowoverwrite: bool = False,
        parallel: str = "ON",
        maxfilesize: Union[str, int, float] = None,
        region: str = None,
    ):
        """Run sql and unload result to S3.

        Args:
            sql: SQL query.
            bucket: S3 bucket name.
            key (optional): S3 key. Create if does not exist.
            prefix (optional): Prefix of the set of files.
            iam_role (optional): IAM Role string. If provided, this will be used as authorization 
                instead of access_key_id/secret_access_key. This feature is untested.
            file_format (optional): CSV or PARQUET.
            manifest (optional): Whether or not to create the manifest file.
            header (optional): Whether or not to include header.
            delimiter (optional): Delimiter charater if the output file is delimited.
            fixedwidth (optional): If not None, it will overwrite delimiter and use fixedwidth 
                format instead.
            encrypted (optional): Whether or not the files should be encrypted.
            bzip2 (optional): Whether or not the files should be compressed with bzip2.
            gzip (optional): Whether or not the files should be compressed with gzip.
            zstd (optional): Whether or not the files should be compressed with zstd.
            addquotes (optional): Whether or not values with delimiter characters should be quoted.
            null (optional): Specify the NULL AS string.
            escape (optional): Whether to include the ESCAPE argument in UNLOAD.
            allowoverwrite (optional): Whether or not existing files should be overwritten. Redshift
                will fail with error message if this is False and there are existing files.
            parallel (optional): ON or OFF. Whether or not to use parallel and unload into multiple 
                files.
            maxfilesize (optional): Maxfilesize argument for UNLOAD.
            region (optional): AWS region if S3 region is different from Redshift region.
        """
        destination_option = ""
        if path is not None:
            destination_option = make_valid_uri(destination_option, f"{path}")
            if destination_option[-1] != "/":
                destination_option = destination_option + "/"
        if prefix is not None:
            destination_option = make_valid_uri(destination_option, f"{prefix}")
        dest_bucket: Optional[str] = bucket or self.default_bucket
        if dest_bucket is None:
            raise ValueError("bucket cannot be None.")
        existing_keys = self._get_s3_pattern_existence(dest_bucket, destination_option)
        if existing_keys:
            warn_message = f"""\
            These keys already exist. It may cause data consistency issues.
            {existing_keys}
            """
            warnings.warn(dedent(warn_message))
        destination_option = make_valid_uri(f"s3://{dest_bucket}", destination_option)

        if sum([bzip2, gzip, zstd]) > 1:
            raise ValueError("Only one of [bzip2, gzip, zstd] should be True.")

        file_format_option = ""
        if file_format is not None:
            if file_format == "CSV":
                if fixedwidth:
                    raise ValueError(
                        "fixedwidth should not be specified for CSV file_format."
                    )
                delimiter = ","
                file_format_option = "format CSV"
            elif file_format == "PARQUET":
                file_format_option = "format PARQUET"
                if delimiter:
                    raise ValueError(
                        "delimiter should not be specified for PARQUET file_format."
                    )
                if fixedwidth:
                    raise ValueError(
                        "fixedwidth should not be specified for PARQUET file_format."
                    )
                if addquotes:
                    raise ValueError(
                        "addquotes should not be specified for PARQUET file_format."
                    )
                if escape:
                    raise ValueError(
                        "escape should not be specified for PARQUET file_format."
                    )
                if null:
                    raise ValueError(
                        "null should not be specified for PARQUET file_format."
                    )
                if header:
                    raise ValueError(
                        "header should not be specified for PARQUET file_format."
                    )
                if gzip:
                    raise ValueError(
                        "gzip should not be specified for PARQUET file_format."
                    )
                if bzip2:
                    raise ValueError(
                        "bzip2 should not be specified for PARQUET file_format."
                    )
                if zstd:
                    raise ValueError(
                        "zstd should not be specified for PARQUET file_format."
                    )
            else:
                raise ValueError("File format can only be CSV or PARQUET if specified.")

        partition_include_option = " INCLUDE" if include_partition_column else ""
        partition_option = (
            f"{','.join(partition_by)}{partition_include_option}"
            if partition_by
            else ""
        )

        manifest_option = "manifest" if manifest else ""
        header_option = "header" if header else ""
        delimiter_option = f"delimiter '{delimiter}'" if delimiter else "delimiter '|'"
        if fixedwidth is not None:
            fixedwidth_option = f"fixedwidth '{fixedwidth}'"
            delimiter_option = ""
        else:
            fixedwidth_option = ""
        encrypted_option = "encrypted" if encrypted else ""
        bzip2_option = "bzip2" if bzip2 else ""
        gzip_option = "gzip" if gzip else ""
        zstd_option = "zstd" if zstd else ""
        addquotes_option = "addquotes" if addquotes else ""
        null_option = f"null as '{null}'" if null is not None else ""
        escape_option = "escape" if escape else ""
        allowoverwrite_option = "allowoverwrite" if allowoverwrite else ""
        parallel_option = f"parallel {parallel}"
        maxfilesize_option = (
            f"maxfilesize '{maxfilesize}'" if maxfilesize is not None else ""
        )
        region_option = f"region {region}" if region else ""
        aws_access_key_id = self.aws_config.get("aws_access_key_id")
        aws_secret_access_key = self.aws_config.get("aws_secret_access_key")
        if (
            aws_access_key_id is None
            and aws_secret_access_key is None
            and iam_role is None
        ):
            raise ValueError(
                "Must provide at least one of [iam_role, aws_access_key_id/aws_secret_access_key]"
            )
        aws_token = self.aws_config.get("aws_session_token")
        aws_token_option = (
            f"session_token '{aws_token}'" if aws_token is not None else ""
        )
        if iam_role is not None:
            iam_role_option = f"iam_role '{iam_role}'"
            access_key_id_option = ""
            secret_access_key_option = ""
        else:
            iam_role_option = ""
            access_key_id_option = f"access_key_id '{aws_access_key_id}'"
            secret_access_key_option = f"secret_access_key '{aws_secret_access_key}'"

        sql = sql.replace("\n", " ")
        unload_template = f"""\
        unload ('{sql}')
        to '{destination_option}'
        {file_format_option}
        {partition_option}
        {manifest_option}
        {header_option}
        {delimiter_option}
        {fixedwidth_option}
        {encrypted_option}
        {bzip2_option}
        {gzip_option}
        {zstd_option}
        {addquotes_option}
        {null_option}
        {escape_option}
        {allowoverwrite_option}
        {parallel_option}
        {maxfilesize_option}
        {region_option}
        {access_key_id_option}
        {secret_access_key_option}
        {aws_token_option}
        {iam_role_option}
        """
        self.run_query(unload_template)



import os
import re
from copy import deepcopy


def filter_kwargs(full, ref):
    return {k: v for k, v in full.items() if k in ref}


def prettify_sql(sql):
    sql = re.sub(r"(?<=access_key_id \')(.*)(?=\')", "*" * 8, sql)
    sql = re.sub(r"(?<=secret_access_key \')(.*)(?=\')", "*" * 8, sql)
    sql = re.sub(r"(?<=iam_role \')(.*)(?=\')", "*" * 8, sql)
    return re.sub(r"\n\s*\n*", "\n", sql.lstrip())


def make_valid_uri(*args):
    if len(args) >= 2:
        l = deepcopy(list(args))
        i = l[1]
        if i[0] == "/":
            l[1] = i[1:]
    else:
        raise ValueError("Need at least two elements to make valid URI.")
    return os.path.join(*l)



PANDAS_TOCSV_KWARGS = [
    "path_or_buf",
    "sep",
    "na_rep",
    "float_format",
    "columns",
    "header",
    "index",
    "index_label",
    "mode",
    "encoding",
    "compression",
    "quoting",
    "quotechar",
    "line_terminator",
    "chunksize",
    "tupleize_cols",
    "date_format",
    "doublequote",
    "escapechar",
    "decimal",
]

PANDAS_READ_TABLE_KWARGS = [
    "sep",
    "delimiter",
    "header",
    "names",
    "index_col",
    "usecols",
    "squeeze",
    "prefix",
    "mangle_dupe_cols",
    "dtype",
    "engine",
    "converters",
    "true_values",
    "false_values",
    "skipinitialspace",
    "skiprows",
    "nrows",
    "na_values",
    "keep_default_na",
    "na_filter",
    "verbose",
    "skip_blank_lines",
    "parse_dates",
    "infer_datetime_format",
    "keep_date_col",
    "date_parser",
    "dayfirst",
    "iterator",
    "chunksize",
    "compression",
    "thousands",
    "decimal",
    "lineterminator",
    "quotechar",
    "quoting",
    "escapechar",
    "comment",
    "encoding",
    "dialect",
    "tupleize_cols",
    "error_bad_lines",
    "warn_bad_lines",
    "skipfooter",
    "doublequote",
    "delim_whitespace",
    "low_memory",
    "memory_map",
    "float_precision",
]



import warnings
from io import StringIO, BytesIO
import logging

import pandas as pd
import boto3

from red_panda.pandas import PANDAS_TOCSV_KWARGS, PANDAS_READ_TABLE_KWARGS
from red_panda.aws import (
    S3_PUT_KWARGS,
    S3_GET_KWARGS,
    S3_CREATE_BUCKET_KWARGS,
)
from red_panda.utils import filter_kwargs, make_valid_uri
from red_panda.aws import AWSUtils


LOGGER = logging.getLogger(__name__)


class S3Utils(AWSUtils):
    """AWS S3 operations.

    Args:
        aws_config: AWS configuration.

    Attributes:
        aws_config (dict): AWS configuration.
    """

    def __init__(self, aws_config: dict):
        super().__init__(aws_config=aws_config)

    def _connect_s3(self):
        """Get S3 session.

        If key/secret are not provided, boto3's default behavior is falling back to awscli configs
        and environment variables.
        """
        return boto3.resource(
            "s3",
            aws_access_key_id=self.aws_config.get("aws_access_key_id"),
            aws_secret_access_key=self.aws_config.get("aws_secret_access_key"),
            aws_session_token=self.aws_config.get("aws_session_token"),
        )

    def _check_s3_bucket_existence(self, bucket: str) -> bool:
        s3 = self.get_s3_client()
        try:
            s3.head_bucket(Bucket=bucket)
        except s3.exceptions.ClientError:
            LOGGER.warning(f"{bucket} does not exist or you do not have access to it.")
            return False
        else:
            return True

    def _check_s3_key_existence(self, bucket: str, key: str) -> bool:
        s3 = self.get_s3_client()
        try:
            s3.head_object(Bucket=bucket, Key=key)
        except s3.exceptions.ClientError:
            LOGGER.warning(
                f"{bucket}/{key} does not exist or you do not have access to it."
            )
            return False
        else:
            return True

    def _get_s3_pattern_existence(self, bucket: str, pattern: str) -> list:
        s3 = self.get_s3_resource()
        all_keys = [
            o.key for o in s3.Bucket(bucket).objects.all() if o.key.startswith(pattern)
        ]
        return all_keys

    def get_s3_resource(self):
        """Return a boto3 S3 resource"""
        return self._connect_s3()

    def get_s3_client(self):
        """Return a boto3 S3 client"""
        return self._connect_s3().meta.client

    def list_buckets(self) -> list:
        """List all buckets.
        
        Returns:
            All S3 buckets for the account.
        """
        s3 = self.get_s3_client()
        response = s3.list_buckets()
        buckets = [bucket["Name"] for bucket in response["Buckets"]]
        return buckets

    def list_object_keys(self, bucket: str, prefix: str = "") -> list:
        """List all object keys.

        Args:
            bucket: Bucket name.
            prefix: Any prefix for the object.

        Returns:
            A list of all objects in a bucket given certain prefix.
        """
        s3 = self.get_s3_client()
        response = s3.list_objects(Bucket=bucket, Prefix=prefix)
        return [o["Key"] for o in response["Contents"]]

    def create_bucket(self, bucket: str, error: str = "warn", **kwargs):
        """Check and create bucket.

        Args:
            bucket: S3 bucket name.
            error (optional): Specify `warn` or `raise` or `silent`. How to handle if bucket already
                exists. Default is `warn`.
            **kwargs: Additional keyword arguments for creating bucket.

        Returns:
            The response from `boto3.create_bucket`.
        """
        s3 = self.get_s3_client()
        if self._check_s3_bucket_existence(bucket):
            if error == "raise":
                raise ValueError(f"{bucket} already exists")
            elif error == "warn":
                warnings.warn(f"{bucket} already exists")
        extra_kwargs = filter_kwargs(kwargs, S3_CREATE_BUCKET_KWARGS)
        return s3.create_bucket(Bucket=bucket, **extra_kwargs)

    def file_to_s3(self, file_name: str, bucket: str, key: str, **kwargs):
        """Put a file to S3.

        Args:
            file_name: Local file name.
            bucket: S3 bucket name.
            key: S3 key.
            **kwargs: ExtraArgs for `boto3.client.upload_file`.
        """
        s3 = self._connect_s3()
        self._check_s3_bucket_existence(bucket)
        s3_put_kwargs = filter_kwargs(kwargs, S3_PUT_KWARGS)
        s3.meta.client.upload_file(
            file_name, Bucket=bucket, Key=key, ExtraArgs=s3_put_kwargs
        )

    def df_to_s3(self, df: pd.DataFrame, bucket: str, key: str, **kwargs):
        """Put DataFrame to S3.

        Args:
            df: Source dataframe.
            bucket: S3 bucket name.
            key: S3 key.
            **kwargs: kwargs for `boto3.Bucket.put_object` and `pandas.DataFrame.to_csv`.
        """
        s3 = self._connect_s3()
        buffer = StringIO()
        to_csv_kwargs = filter_kwargs(kwargs, PANDAS_TOCSV_KWARGS)
        df.to_csv(buffer, **to_csv_kwargs)
        self._check_s3_bucket_existence(bucket)
        s3_put_kwargs = filter_kwargs(kwargs, S3_PUT_KWARGS)
        s3.Bucket(bucket).put_object(Key=key, Body=buffer.getvalue(), **s3_put_kwargs)

    def delete_from_s3(self, bucket: str, key: str):
        """Delete object from S3.

        Args:
            bucket: S3 bucket name.
            key: S3 key.
        """
        s3 = self._connect_s3()
        if self._check_s3_key_existence(bucket, key):
            s3.meta.client.delete_object(Bucket=bucket, Key=key)
        else:
            LOGGER.warning(f"{bucket}: {key} does not exist.")

    def delete_bucket(self, bucket: str):
        """Empty and delete bucket.

        Args:
            bucket: S3 bucket name.

        TODO:
            * Handle when there is bucket versioning.
        """
        s3_bucket = self.get_s3_resource().Bucket(bucket)
        s3_bucket.objects.all().delete()
        s3_bucket.delete()

    def s3_to_obj(self, bucket: str, key: str, **kwargs) -> BytesIO:
        """Read S3 object into memory as BytesIO.

        Args:
            bucket: S3 bucket name.
            key: S3 key.
            **kwargs: kwargs for `boto3.client.get_object`.
        """
        s3_get_kwargs = filter_kwargs(kwargs, S3_GET_KWARGS)
        s3 = self.get_s3_client()
        obj = s3.get_object(Bucket=bucket, Key=key, **s3_get_kwargs)
        return BytesIO(obj["Body"].read())

    def s3_to_file(self, bucket: str, key: str, file_name: str, **kwargs):
        """Download S3 object as local file.

        Args:
            bucket: S3 bucket name.
            key: S3 key.
            file_name: Local file name.
            **kwargs: kwargs for `boto3.client.download_file`.
        """
        s3_get_kwargs = filter_kwargs(kwargs, S3_GET_KWARGS)
        s3 = self.get_s3_resource()
        s3.Bucket(bucket).download_file(
            Key=key, Filename=file_name, ExtraArgs=s3_get_kwargs
        )

    def s3_to_df(self, bucket: str, key: str, **kwargs):
        """Read S3 object into memory as DataFrame

        Only supporting delimited files. Default is tab delimited files.

        Args:
            bucket: S3 bucket name.
            key: S3 key.
            **kwargs: kwargs for `pandas.read_table` and `boto3.client.get_object`.

        Returns:
            A DataFrame.
        """
        s3_get_kwargs = filter_kwargs(kwargs, S3_GET_KWARGS)
        read_table_kwargs = filter_kwargs(kwargs, PANDAS_READ_TABLE_KWARGS)
        buffer = self.s3_to_obj(bucket, key, **s3_get_kwargs)
        return pd.read_csv(buffer, **read_table_kwargs)

    def s3_folder_to_df(self, bucket: str, folder: str, prefix: str = None, **kwargs):
        """Read all files in folder with prefix to a df.

        Args:
            bucket: S3 bucket name.
            folder: S3 folder.
            prefix: File prefix.

        Returns:
            A DataFrame.
        """
        s3_get_kwargs = filter_kwargs(kwargs, S3_GET_KWARGS)
        read_table_kwargs = filter_kwargs(kwargs, PANDAS_READ_TABLE_KWARGS)
        if folder[-1] != "/":
            folder = folder + "/"
        pattern = make_valid_uri(folder, prefix or "/")
        allfiles = [f for f in self.list_object_keys(bucket, pattern) if f != folder]
        dfs = []
        for f in allfiles:
            LOGGER.info(f"Reading file {f}")
            dfs.append(self.s3_to_df(bucket, f, **s3_get_kwargs, **read_table_kwargs))
        return pd.concat(dfs)



from typing import Tuple, Union
import pandas as pd


QueryResult = Tuple[Union[dict, None], Union[list, None]]
TemplateQueryResult = Union[QueryResult, pd.DataFrame]
AthenaQueryResult = Union[list, pd.DataFrame]



import logging
from typing import Union, Optional, List

import pandas as pd
import psycopg2

from red_panda.typing import QueryResult, TemplateQueryResult
from red_panda.pandas import PANDAS_TOCSV_KWARGS
from red_panda.aws.templates.redshift import (
    SQL_NUM_SLICES,
    SQL_TABLE_INFO,
    SQL_TABLE_INFO_SIMPLIFIED,
    SQL_LOAD_ERRORS,
    SQL_RUNNING_INFO,
    SQL_LOCK_INFO,
    SQL_TRANSACT_INFO,
)
from red_panda.utils import filter_kwargs, prettify_sql


LOGGER = logging.getLogger(__name__)


def create_column_definition_single(d: dict) -> str:
    """Create the column definition for a single column.

    Args:
        d: A `dict` of values to compose a single column definition, defaults::

                {
                    "data_type": "varchar(256)", # str
                    "default": None, # Any
                    "identity": None, # tuple
                    "encode": None, # str
                    "distkey": False,
                    "sortkey": False,
                    "nullable": True,
                    "unique": False,
                    "primary_key": False,
                    "foreign_key": False,
                    "references": None, # str
                    "like": None, # str
                }
    
    Returns:
        str: Single column definition for Redshift.
    """
    data_type = d.get("data_type")
    data_type_option = data_type if data_type is not None else "varchar(256)"
    default = d.get("default")
    quote = "'" if not isinstance(default, (int, float, complex)) else ""
    default_option = f"default {quote}{default}{quote}" if default is not None else ""
    identity = d.get("identity")
    if identity is not None:
        seed, step = identity
        identity_option = f"identity({seed}, {step})"
    else:
        identity_option = ""
    encode = d.get("encode")
    encode_option = f"encode {encode}" if encode is not None else ""
    distkey = d.get("distkey")
    distkey_option = "distkey" if distkey is not None and distkey else ""
    sortkey = d.get("sortkey")
    sortkey_option = "sortkey" if sortkey is not None and sortkey else ""
    nullable = d.get("nullable")
    nullable_option = "not null" if nullable is not None and not nullable else ""
    unique = d.get("unique")
    unique_option = "unique" if unique is not None and unique else ""
    primary_key = d.get("primary_key")
    primary_key_option = (
        "primary key" if primary_key is not None and primary_key else ""
    )
    references = d.get("references")
    references_option = f"references {references}" if references is not None else ""
    like = d.get("like")
    like_option = f"like {like}" if like is not None else ""
    template = " ".join(
        [
            data_type_option,
            default_option,
            identity_option,
            encode_option,
            distkey_option,
            sortkey_option,
            nullable_option,
            unique_option,
            primary_key_option,
            references_option,
            like_option,
        ]
    )
    return " ".join(template.split())


def create_column_definition(d: dict) -> str:
    """Create full column definition string for Redshift.

    Args:
        d: A `dict` if single column definitions, where the keys are column names.

    Returns:
        str: Full column definition for Redshift.
    """
    return ",\n".join(f"{c} {create_column_definition_single(o)}" for c, o in d.items())


class RedshiftUtils:
    """ Base class for Redshift operations.

    Args:
        redshift_conf: Redshift configuration.
        dryrun (optional): If True, queries will be printed instead of executed.
    
    Attributes:
        redshift_conf (dict): Redshift configuration.
    """

    def __init__(self, redshift_config: dict, dryrun: bool = False):
        self.redshift_config = redshift_config
        self._dryrun = dryrun

    def _connect_redshift(self):
        return psycopg2.connect(
            user=self.redshift_config.get("user"),
            password=self.redshift_config.get("password"),
            host=self.redshift_config.get("host"),
            port=self.redshift_config.get("port"),
            dbname=self.redshift_config.get("dbname"),
        )

    def run_sql_from_file(self, file_name: str):
        """Run a `.sql` file.

        Args:
            file_name: SQL file to be run.

        TODO:
            * Add support for transactions.
        """
        with open(file_name, "r") as f:
            sql = f.read()
        self.run_query(sql)

    def run_query(self, sql: str, fetch: bool = False) -> QueryResult:
        """Run a SQL query.

        Args:
            sql: SQL string.
            fetch (optional): Whether to return data from the query.

        Returns:
            Tuple[dict, list]: (data, columns) where data is a json/dict representation of the data 
            and columns is a list of column names.
        """
        LOGGER.info(prettify_sql(sql))

        if self._dryrun:
            return (None, None)

        conn = self._connect_redshift()
        cursor = conn.cursor()
        columns = None
        data = None
        try:
            cursor.execute(sql)
            if fetch:
                if cursor.description is not None:
                    columns = [desc[0] for desc in cursor.description]
                    data = cursor.fetchall()
                else:
                    LOGGER.warning("Query completed but it returned no data.")
            else:
                conn.commit()
        except KeyboardInterrupt:
            conn.cancel()
            LOGGER.warning("User canceled query.")
        finally:
            conn.close()
        return (data, columns)

    def cancel_query(self, pid: Union[str, int], transaction: bool = False):
        """Cancels a running query given pid.

        Args:
            pid: PID of a running query in Redshift.
            transaction (optional): Whether the running query is a transaction.
        """
        self.run_query(f"cancel {pid}")
        if transaction:
            self.run_query("abort")

    def kill_session(self, pid: Union[str, int]):
        """ Kill a session given pid.

        Args:
            pid: PID of a running query in Redshift.
        """
        self.run_query(f"select pg_terminate_backend({pid})")

    def get_num_slices(self) -> Optional[int]:
        """Get number of slices of a Redshift cluster.
        
        Returns:
            int: Number of slices of the connected cluster.

        Raises:
            IndexError: When Redshift returns invalid number of slices.
        """
        data, _ = self.run_query(SQL_NUM_SLICES, fetch=True)
        n_slices = None
        try:
            n_slices = data[0][0]
        except IndexError:
            LOGGER.error("Could not derive number of slices of Redshift cluster.")
        return n_slices

    def run_template(self, sql: str, as_df: bool = True) -> TemplateQueryResult:
        """Utility method to run a pre-defined sql template.

        Args:
            sql: SQL string.
            as_df (optional): Whether or not to return the result as a `pandas.DataFrame`.

        Returns:
            Union[pandas.DataFrame, Tuple[dict, list]]: Either return a DataFrame/table of the 
            template query result or the raw form as specified in `run_query`.
        """
        data, columns = self.run_query(sql, fetch=True)
        if as_df:
            return pd.DataFrame(data, columns=columns)
        else:
            return (data, columns)

    def get_table_info(
        self, as_df: bool = True, simple: bool = False
    ) -> TemplateQueryResult:
        """Utility to get table information in the cluster.

        Args:
            as_df (optional): Whether or not to return the result as a `pandas.DataFrame`.
            simple (optional): Whether to get the basic table information only.

        Returns:
            Table information.
        """
        sql = SQL_TABLE_INFO_SIMPLIFIED if simple else SQL_TABLE_INFO
        return self.run_template(sql, as_df)

    def get_load_error(self, as_df: bool = True) -> TemplateQueryResult:
        """Utility to get load errors in the cluster.

        Args:
            as_df (optional): Whether or not to return the result as a `pandas.DataFrame`.

        Returns:
            Load error information.
        """
        return self.run_template(SQL_LOAD_ERRORS, as_df)

    def get_running_info(self, as_df: bool = True) -> TemplateQueryResult:
        """Utility to get information on running queries in the cluster.

        Args:
            as_df (optional): Whether or not to return the result as a `pandas.DataFrame`.

        Returns:
            Running query information.
        """
        return self.run_template(SQL_RUNNING_INFO, as_df)

    def get_lock_info(self, as_df: bool = True) -> TemplateQueryResult:
        """Utility to get lock information in the cluster.

        Args:
            as_df (optional): Whether or not to return the result as a `pandas.DataFrame`.

        Returns:
            Lock information.
        """
        return self.run_template(SQL_LOCK_INFO, as_df)

    def get_transaction_info(self, as_df: bool = True) -> TemplateQueryResult:
        """Utility to get transaction information in the cluster.

        Args:
            as_df (optional): Whether or not to return the result as a `pandas.DataFrame`.

        Returns:
            Transaction information.
        """
        return self.run_template(SQL_TRANSACT_INFO, as_df)

    def redshift_to_df(self, sql: str) -> pd.DataFrame:
        """Redshift query result to a Pandas DataFrame.

        Args:
            sql: SQL query.

        Returns:
            pandas.DataFrame: A DataFrame of query result.
        """
        data, columns = self.run_query(sql, fetch=True)
        return pd.DataFrame(data, columns=columns)

    def redshift_to_file(self, sql: str, file_name: str, **kwargs):
        """Redshift query result to a file.

        Args:
            sql: SQL query.
            file_name: File name of the saved file.
            **kwargs: `to_csv` keyword arguments.
        """
        data = self.redshift_to_df(sql)
        to_csv_kwargs = filter_kwargs(kwargs, PANDAS_TOCSV_KWARGS)
        data.to_csv(file_name, **to_csv_kwargs)

    def create_table(
        self,
        table_name: str,
        column_definition: dict,
        temp: bool = False,
        if_not_exists: bool = False,
        backup: str = "YES",
        unique: List[str] = None,
        primary_key: str = None,
        foreign_key: List[str] = None,
        references: List[str] = None,
        diststyle: str = "EVEN",
        distkey: str = None,
        sortstyle: str = "COMPOUND",
        sortkey: List[str] = None,
        drop_first: bool = False,
    ):
        """Utility for creating a table in Redshift.

        Args:
            table_name: Name of table to be created.
            column_definition: default::

                {
                    "col1": {
                        "data_type": "varchar(256)", # str
                        "default": None, # Any
                        "identity": None, # tuple(int, int)
                        "encode": None, # str
                        "distkey": False,
                        "sortkey": False,
                        "nullable": True,
                        "unique": False,
                        "primary_key": False,
                        "foreign_key": False,
                        "references": None, # str
                        "like": None, # str
                    },
                    "col2": {
                        ...
                    },
                    ...
                }
            temp (optional): Corresponding argument of `create table` in Redshift.  
            if_not_exists (optional): Corresponding argument of `create table` in Redshift.  
            backup (optional): Corresponding argument of `create table` in Redshift.  
            unique (optional): Corresponding argument of `create table` in Redshift.  
            primary_key (optional): Corresponding argument of `create table` in Redshift.  
            foreign_key (optional): Corresponding argument of `create table` in Redshift.  
                Must match references.
            references (optional): Corresponding argument of `create table` in Redshift.  
                Must match foreign_key.
            diststyle (optional): Corresponding argument of `create table` in Redshift.  
            distkey (optional): Corresponding argument of `create table` in Redshift.  
            sortstyle (optional): Corresponding argument of `create table` in Redshift.  
            sortkey (optional): Corresponding argument of `create table` in Redshift.  
            drop_first (optional): Corresponding argument of `create table` in Redshift.  

        TODO:
            * Check consistency between column_constraints and table_constraints
        """
        if drop_first:
            self.run_query(f"drop table if exists {table_name}")
        temp_option = "temp" if temp else ""
        exist_option = "if not exists" if if_not_exists else ""
        unique_option = f'unique ({", ".join(unique)})' if unique is not None else ""
        primary_key_option = (
            f"primary key ({primary_key})" if primary_key is not None else ""
        )
        foreign_key_option = (
            f'foreign key ({", ".join(foreign_key)})' if foreign_key is not None else ""
        )
        references_option = (
            f'references ({", ".join(references)})' if references is not None else ""
        )
        distkey_option = f"distkey({distkey})" if distkey is not None else ""
        sortkey_option = (
            f'{sortstyle} sortkey({", ".join(sortkey)})' if sortkey is not None else ""
        )
        create_template = f"""\
        create table {temp_option} {table_name} {exist_option} (
        {create_column_definition(column_definition)}
        )
        backup {backup}
        diststyle {diststyle}
        {unique_option}
        {primary_key_option}
        {foreign_key_option}
        {references_option}
        {distkey_option}
        {sortkey_option}
        """
        self.run_query(create_template)



REDSHIFT_RESERVED_WORDS = [
    "aes128",
    "aes256",
    "all",
    "allowoverwrite",
    "analyse",
    "analyze",
    "and",
    "any",
    "array",
    "as",
    "asc",
    "authorization",
    "backup",
    "between",
    "binary",
    "blanksasnull",
    "both",
    "bytedict",
    "bzip2",
    "case",
    "cast",
    "check",
    "collate",
    "column",
    "constraint",
    "create",
    "credentials",
    "cross",
    "current_date",
    "current_time",
    "current_timestamp",
    "current_user",
    "current_user_id",
    "default",
    "deferrable",
    "deflate",
    "defrag",
    "delta",
    "delta32k",
    "desc",
    "disable",
    "distinct",
    "do",
    "else",
    "emptyasnull",
    "enable",
    "encode",
    "encrypt     ",
    "encryption",
    "end",
    "except",
    "explicit",
    "false",
    "for",
    "foreign",
    "freeze",
    "from",
    "full",
    "globaldict256",
    "globaldict64k",
    "grant",
    "group",
    "gzip",
    "having",
    "identity",
    "ignore",
    "ilike",
    "in",
    "initially",
    "inner",
    "intersect",
    "into",
    "is",
    "isnull",
    "join",
    "leading",
    "left",
    "like",
    "limit",
    "localtime",
    "localtimestamp",
    "lun",
    "luns",
    "lzo",
    "lzop",
    "minus",
    "mostly13",
    "mostly32",
    "mostly8",
    "natural",
    "new",
    "not",
    "notnull",
    "null",
    "nulls",
    "off",
    "offline",
    "offset",
    "oid",
    "old",
    "on",
    "only",
    "open",
    "or",
    "order",
    "outer",
    "overlaps",
    "parallel",
    "partition",
    "percent",
    "permissions",
    "placing",
    "primary",
    "raw",
    "readratio",
    "recover",
    "references",
    "respect",
    "rejectlog",
    "resort",
    "restore",
    "right",
    "select",
    "session_user",
    "similar",
    "snapshot ",
    "some",
    "sysdate",
    "system",
    "table",
    "tag",
    "tdes",
    "text255",
    "text32k",
    "then",
    "timestamp",
    "to",
    "top",
    "trailing",
    "true",
    "truncatecolumns",
    "union",
    "unique",
    "user",
    "using",
    "verbose",
    "wallet",
    "when",
    "where",
    "with",
    "without",
]

REDSHIFT_COPY_KWARGS = [
    "delimiter",
    "ignoreheader",
    "quote_character",
    "dateformat",
    "timeformat",
    "region",
    "null",
    "escape",  # bool
    "acceptinvchars",  # str
    "iam_role",
    "acceptanydate",  # bool
    "column_list",  # list[str]
    "blanksasnull",  # bool
    "emptyasnull",  # bool
    "encoding",  # str
    "explicit_ids",  # bool
    "fillrecord",  # bool
    "ignoreblanklines",  # bool
    "removequotes",  # bool
    "roundec",  # bool
    "trimblanks",  # bool
    "truncatecolumns",  # bool
]

REDSHIFT_UNLOAD_KWARGS = [
    "manifest",  # bool
    "delimiter",
    "fixedwidth",
    "encrypted",  # bool
    "bzip2",  # bool
    "gzip",  # bool
    "addquotes",  # bool
    "null",
    "escap",  # bool
    "allowoverwrite",  # bool
    "parallel",
    "maxfilesize",
]

S3_PUT_KWARGS = [
    "ACL",
    # 'Body', required in our context and handled seperately
    "CacheControl",
    "ContentDisposition",
    "ContentEncoding",
    "ContentLanguage",
    "ContentLength",
    "ContentMD5",
    "ContentType",
    "Expires",
    "GrantFullControl",
    "GrantRead",
    "GrantReadACP",
    "GrantWriteACP",
    # 'Key', required and handled seperately
    "Metadata",
    "ServerSideEncryption",
    "StorageClass",
    "WebsiteRedirectLocation",
    "SSECustomerAlgorithm",
    "SSECustomerKey",
    "SSEKMSKeyId",
    "RequestPayer",
    "Tagging",
]

S3_GET_KWARGS = [
    "Bucket",
    "IfMatch",
    "IfModifiedSince",
    "IfNoneMatch",
    "IfUnmodifiedSince",
    "Key",
    "Range",
    "ResponseCacheControl",
    "ResponseContentDisposition",
    "ResponseContentEncoding",
    "ResponseContentLanguage",
    "ResponseContentType",
    "ResponseExpires",
    "VersionId",
    "SSECustomerAlgorithm",
    "SSECustomerKey",
    "RequestPayer",
    "PartNumber",
]

S3_CREATE_BUCKET_KWARGS = [
    "ACL",
    "CreateBucketConfiguration",
    "GrantFullControl",
    "GrantRead",
    "GrantReadACP",
    "GrantWrite",
    "GrantWriteACP",
]


class AWSUtils:
    """ Base class for AWS operations.
    
    Args:
        aws_config: AWS configuration.

    Attributes:
        aws_config (dict): AWS configuration.
    """

    def __init__(self, aws_config: dict):
        if aws_config is None:
            aws_config = {
                "aws_access_key_id": None,
                "aws_secret_access_key": None,
                "aws_session_token": None,
                "region_name": None,
            }
        self.aws_config = aws_config



from functools import reduce
from typing import Union, List, Dict, Callable
import pandas as pd


def merge_dfs(dfs: List[pd.DataFrame], **kwargs) -> pd.DataFrame:
    """Merge a list of DataFrames on common columns.
    
    Args:
        dfs: A list of `pandas.DataFrame`s.
        **kwargs: Keyword arguments for `pandas.merge`.
    
    Returns:
        Merged DataFrame.
    """
    return reduce(lambda df1, df2: pd.merge(df1, df2, **kwargs), dfs)


def row_number(
    df: pd.DataFrame, group_by: List[str], sort_by: List[str], ascending: bool = True,
) -> pd.Series:
    """Create a row number series given a DataFrame lists of columns for group by and sort by.
    
    Args:
        df: Input DataFrame.
        group_by: List of group by columns.
        sort_by: List of sort by columns.
        col_name (optional): The output column name.
        ascending (optional): Whether sort in ascending order.
        as_series (optional): Whether to return a Series instead of a DataFrame.

    Returns:
        A DataFrame with row number or the row number Series.

    Example:
        >>> df = row_number(df, ['group'], ['sort'], as_series=False)
        >>> df['rn'] = row_number(df, ['group'], ['sort'])
    """
    return df.sort_values(sort_by, ascending=ascending).groupby(group_by).cumcount()


def groupby_mutate(
    df: pd.DataFrame,
    group_by: Union[List[str], str],
    func_dict: Dict[str, Callable],
    inplace: bool = False,
) -> pd.DataFrame:
    """Similar to R's dplyr::mutate.

    Example:
        >>> def func(x):
                return x["x"] / sum(x["x"])
        >>> func_dict = {
                'ratio': x["x"] / sum(x["x"])
            }
        >>> groupby_mutate(df, "b", func_dict)
    """
    out = df if inplace else df.copy()
    for col, func in func_dict.items():
        out[col] = out.groupby(group_by, group_keys=False).apply(func)
    return out



SQL_NUM_SLICES = "select count(1) from stv_slices"

SQL_LOAD_ERRORS = "select * from stl_load_errors order by starttime desc"

SQL_TABLE_INFO = """\
SELECT TRIM(pgn.nspname) AS SCHEMA,
TRIM(a.name) AS TABLE,
id AS TableId,
decode(pgc.reldiststyle, 0, 'EVEN', 1, det.distkey, 8, 'ALL') AS DistKey,
decode(pgc.reldiststyle,8, NULL, dist_ratio.ratio::DECIMAL(20,4)) AS Skew,
det.head_sort AS "SortKey",
det.n_sortkeys AS "#SKs",
CASE WHEN pgc.reldiststyle = 8 THEN a.rows_all_dist ELSE a.rows END AS rows,
b.mbytes,
decode(det.max_enc, 0, 'N', 'Y') AS Enc, 
det.pct_enc, 
decode(b.mbytes, 0, 0,((b.mbytes/part.total::DECIMAL)*100)::DECIMAL(20,2)) AS pct_of_total,
CASE 
    WHEN a.rows = 0 
        THEN NULL 
    ELSE 
        CASE 
            WHEN pgc.reldiststyle = 8 
                THEN ((a.rows_all_dist - pgc.reltuples)::DECIMAL(20,3) / a.rows_all_dist::DECIMAL(20,3)*100)::DECIMAL(20,2)
            ELSE ((a.rows - pgc.reltuples)::DECIMAL(20,3) / a.rows::DECIMAL(20,3)*100)::DECIMAL(20,2) 
    END 
END AS pct_stats_off,
CASE 
    WHEN pgc.reldiststyle = 8 
        THEN decode(det.n_sortkeys, 0, NULL, DECODE(a.rows_all_dist, 0, 0, (a.unsorted_rows_all_dist::DECIMAL(32)/a.rows_all_dist)*100))::DECIMAL(20,2)
    ELSE decode(det.n_sortkeys, 0, NULL, DECODE(a.rows, 0, 0, (a.unsorted_rows::DECIMAL(32)/a.rows)*100))::DECIMAL(20,2) 
END AS pct_unsorted
FROM (
    SELECT db_id,
    id,
    name,
    SUM(ROWS) AS ROWS,
    MAX(ROWS) AS rows_all_dist,
    SUM(ROWS) - SUM(sorted_rows) AS unsorted_rows,
    MAX(ROWS) - MAX(sorted_rows) AS unsorted_rows_all_dist
    FROM stv_tbl_perm a
    GROUP BY db_id,
    id,
    name
) AS a
JOIN pg_class AS pgc ON pgc.oid = a.id
JOIN pg_namespace AS pgn ON pgn.oid = pgc.relnamespace
LEFT OUTER JOIN (
    SELECT tbl, COUNT(*) AS mbytes FROM stv_blocklist GROUP BY tbl
) b ON a.id = b.tbl
INNER JOIN (
    SELECT attrelid,
    MIN(CASE attisdistkey WHEN 't' THEN attname ELSE NULL END) AS "distkey",
    MIN(CASE attsortkeyord WHEN 1 THEN attname ELSE NULL END) AS head_sort,
    MAX(attsortkeyord) AS n_sortkeys,
    MAX(attencodingtype) AS max_enc,
    SUM(case when attencodingtype <> 0 then 1 else 0 end)::DECIMAL(20,3)/COUNT(attencodingtype)::DECIMAL(20,3) * 100.00 as pct_enc
    FROM pg_attribute
    GROUP BY 1
) AS det ON det.attrelid = a.id
INNER JOIN (
    SELECT tbl,
    MAX(Mbytes)::DECIMAL(32) /MIN(Mbytes) AS ratio
    FROM (
        SELECT tbl,
        TRIM(name) AS name,
        slice,
        COUNT(*) AS Mbytes
        FROM svv_diskusage
        GROUP BY tbl,
        name,
        slice
    )
    GROUP BY tbl,
    name
) AS dist_ratio ON a.id = dist_ratio.tbl
JOIN (
    SELECT SUM(capacity) AS total
    FROM stv_partitions
    WHERE part_begin = 0
) AS part ON 1 = 1
WHERE mbytes IS NOT NULL
AND   pgc.relowner > 1
ORDER BY mbytes DESC
"""

SQL_TABLE_INFO_SIMPLIFIED = """\
select table_catalog, table_schema, table_name, table_type 
from information_schema.tables 
where table_schema not in ('pg_catalog', 'information_schema')
"""

SQL_RUNNING_INFO = """\
select trim(u.usename) as user, 
s.pid, 
q.xid,
q.query,
q.service_class as "q", 
q.slot_count as slt, 
date_trunc('second',q.wlm_start_time) as start,
decode(trim(q.state), 'Running', 'Run', 'QueuedWaiting', 'Queue', 'Returning', 'Return', trim(q.state)) as state, 
q.queue_Time/1000000 as q_sec, 
q.exec_time/1000000 as exe_sec, 
m.cpu_time/1000000 cpu_sec, 
m.blocks_read read_mb, 
decode(m.blocks_to_disk,-1,null,m.blocks_to_disk) spill_mb, 
m2.rows as ret_rows, m3.rows as NL_rows,
substring(replace(nvl(qrytext_cur.text,trim(translate(s.text,chr(10)||chr(13)||chr(9) ,''))),'\\n',' '), 1, 90) as sql,
trim(decode(event&1,1,'SK ','') || decode(event&2,2,'Del ','') || decode(event&4,4,'NL ','') || decode(event&8,8,'Dist ','') || decode(event&16,16,'Bcast ','') || decode(event&32,32,'Stats ','')) as Alert
from stv_wlm_query_state q 
left outer join stl_querytext s on (s.query=q.query and sequence = 0)
left outer join stv_query_metrics m on ( q.query = m.query and m.segment=-1 and m.step=-1 )
left outer join stv_query_metrics m2 on ( q.query = m2.query and m2.step_type = 38 )
left outer join ( select query, sum(rows) as rows from stv_query_metrics m3 where step_type = 15 group by 1) as m3 on ( q.query = m3.query )
left outer join pg_user u on ( s.userid = u.usesysid )
LEFT OUTER JOIN (
    SELECT ut.xid,'CURSOR ' || TRIM( substring ( TEXT from strpos(upper(TEXT),'SELECT') )) as TEXT
    FROM stl_utilitytext ut
    WHERE sequence = 0
    AND upper(TEXT) like 'DECLARE%'
    GROUP BY text, ut.xid) qrytext_cur ON 
(q.xid = qrytext_cur.xid)
left outer join (
    select query,
    sum(decode(
        trim(split_part(event,':',1)), 'Very selective query filter', 
        1, 'Scanned a large number of deleted rows', 
        2, 'Nested Loop Join in the query plan',
        4,'Distributed a large number of rows across the network',
        8,'Broadcasted a large number of rows across the network',
        16,'Missing query planner statistics',32,0)
    ) as event 
    from STL_ALERT_EVENT_LOG 
    where event_time >=  dateadd(hour, -8, current_Date) 
    group by query  
) as alrt 
on alrt.query = q.query
order by q.service_class,q.exec_time desc, q.wlm_start_time
"""

SQL_LOCK_INFO = """\
select table_id, 
last_update, 
last_commit, 
lock_owner_pid, 
lock_status 
from stv_locks 
order by last_update asc
"""

SQL_TRANSACT_INFO = """\
select * from svv_transactions
"""



import io
import re

from collections import OrderedDict
from setuptools import setup, find_packages


with open("README.md") as f:
    readme = f.read()

with io.open("red_panda/__init__.py", "rt") as f:
    version = re.search(r"__version__ = \"(.*?)\"", f.read()).group(1)

setup(
    name="red-panda",
    version=version,
    url="https://github.com/jucyai/red-panda",
    project_urls=OrderedDict(
        (
            ("Code", "https://github.com/jucyai/red-panda"),
            ("Issue tracker", "https://github.com/jucyai/red-panda/issues"),
        )
    ),
    license="MIT",
    author="Jiachen Yao",
    maintainer="Jiachen Yao",
    description="Data science on the cloud",
    long_description_content_type="text/markdown",
    long_description=readme,
    packages=find_packages(exclude=["tests", "cdk"]),
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "pandas>=1.1.0",
        "psycopg2-binary>=2.8.5",
        "boto3>=1.14.38",
        "awscli>=1.18.115",
        "PyAthena>=1.11.0"
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
